@@ -62,6 +62,8 @@
 #include <txt/txt.h>
 #include <txt/heap.h>
 
+#define MAJOR_VER(v)      ((v) >> 8)
+
 extern void shutdown(void);
 extern void s3_launch(void);
 
@@ -135,6 +137,36 @@ static uint8_t _policy_index_buf[MAX_POLICY_SIZE];
 
 /* default policy */
 static const tb_policy_t _def_policy = {
+    version        : 2,
+    policy_type    : TB_POLTYPE_CONT_NON_FATAL,
+    hash_alg       : TB_HALG_SHA256,
+    policy_control : TB_POLCTL_EXTEND_PCR17,
+    num_entries    : 3,
+    entries        : {
+        {   /* mod 0 is extended to PCR 18 by default, so don't re-extend it */
+            mod_num    : 0,
+            pcr        : TB_POL_PCR_NONE,
+            hash_type  : TB_HTYPE_ANY,
+            num_hashes : 0
+        },
+        {   /* all other modules are extended to PCR 19 */
+            mod_num    : TB_POL_MOD_NUM_ANY,
+            pcr        : 19,
+            hash_type  : TB_HTYPE_ANY,
+            num_hashes : 0
+        },
+        {   /* NV index for geo-tagging will be extended to PCR 22 */
+            mod_num    : TB_POL_MOD_NUM_NV_RAW,
+            pcr        : 22,
+            hash_type  : TB_HTYPE_ANY,
+            nv_index   : 0x40000010,
+            num_hashes : 0
+        }
+    }
+};
+
+/* default policy for TPM 1.2 */
+static const tb_policy_t _def_policy_12 = {
     version        : 2,
     policy_type    : TB_POLTYPE_CONT_NON_FATAL,
     hash_alg       : TB_HALG_SHA1,
@@ -167,7 +199,7 @@ static const tb_policy_t _def_policy = {
 static const tb_policy_t _def_policy_da = {
     version        : 2,
     policy_type    : TB_POLTYPE_CONT_NON_FATAL,
-    hash_alg       : TB_HALG_SHA1,
+    hash_alg       : TB_HALG_SHA256,
     policy_control : TB_POLCTL_EXTEND_PCR17,
     num_entries    : 3,
     entries        : {
@@ -254,6 +286,24 @@ static bool read_policy_from_tpm(uint32_t index, void* policy_index, size_t *pol
     return true;
 }
 
+bool check_index_attribute(uint32_t index)
+{
+    uint32_t attribute;
+    struct tpm_if *tpm = get_tpm();
+    const struct tpm_if_fp *tpm_fp = get_tpm_fp();
+
+    if ( !tpm_fp->get_nvindex_permission(tpm, 0, index, &attribute) ) {
+        printk(TBOOT_ERR"\treading nv index permission failed\n");
+        return false;
+    }
+    if (attribute & TPM_NV_PER_PPWRITE) {
+        printk(TBOOT_ERR"\tnv index should not be TPMA_NV_PPWRITE\n");
+        return false;
+    }
+
+    return true;
+}
+
 /*
  * unwrap_lcp_policy
  *
@@ -263,7 +313,9 @@ static bool read_policy_from_tpm(uint32_t index, void* policy_index, size_t *pol
 static bool unwrap_lcp_policy(void)
 {
     void* lcp_base;
+    lcp_policy_element_t *elt;
     uint32_t lcp_size;
+    uint16_t list_ver;
 
     // scaffolding
     printk(TBOOT_INFO"in unwrap_lcp_policy\n");
@@ -282,14 +334,38 @@ static bool unwrap_lcp_policy(void)
     }
 
     /* if lcp policy data version is 2+ */
-    if ( memcmp((void *)lcp_base, LCP_POLICY_DATA_FILE_SIGNATURE,
+    if ( tb_memcmp((void *)lcp_base, LCP_POLICY_DATA_FILE_SIGNATURE,
              LCP_FILE_SIG_LENGTH) == 0 ) {
-        lcp_policy_data_t *poldata = (lcp_policy_data_t *)lcp_base;
-        lcp_policy_list_t *pollist = &poldata->policy_lists[0];
+        uint32_t elts_size;
+        uint32_t all_elt_size; //Value from lcp list PolicyElementsSize field
+        lcp_policy_data_t2 *poldata = (lcp_policy_data_t2 *)lcp_base;
+        lcp_list_t *pollist = &poldata->policy_lists[0];
 
+        //Read list version
+        tb_memcpy((void*)&list_ver, (const void *)pollist, sizeof(uint16_t));
         for ( int i = 0; i < poldata->num_lists; i++ ) {
-            lcp_policy_element_t *elt = pollist->policy_elements;
-            uint32_t elts_size = 0;
+            if (MAJOR_VER(list_ver) == MAJOR_VER(LCP_TPM12_POLICY_LIST_VERSION)) {
+                elt = pollist->tpm12_policy_list.policy_elements;
+                elts_size = 0;
+                all_elt_size = pollist->tpm12_policy_list.policy_elements_size;
+            }
+            else if (MAJOR_VER(list_ver) ==
+                        MAJOR_VER(LCP_TPM20_POLICY_LIST_VERSION)) {
+                elt = pollist->tpm20_policy_list.policy_elements;
+                elts_size = 0;
+                all_elt_size = pollist->tpm20_policy_list.policy_elements_size;
+            }
+            else if (MAJOR_VER(list_ver) ==
+                        MAJOR_VER(LCP_TPM20_POLICY_LIST2_1_VERSION_300)) {
+                elt = pollist->tpm20_policy_list_2_1.PolicyElements;
+                elts_size = 0;
+                all_elt_size = pollist->tpm20_policy_list_2_1.PolicyElementsSize;
+            }
+            else {
+                printk(TBOOT_ERR"Error: Policy list version not recognized: 0x%x.\n",
+                        list_ver);
+                return false;
+            }
 
             while ( elt ) {
                 /* check element type */
@@ -301,23 +377,31 @@ static bool unwrap_lcp_policy(void)
                     /* check uuid in custom element */
                     if ( are_uuids_equal(&custom->uuid,
                              &((uuid_t)LCP_CUSTOM_ELEMENT_TBOOT_UUID)) ) {
-                        memcpy(_policy_index_buf, &custom->data,
+                        tb_memcpy(_policy_index_buf, &custom->data,
                             elt->size - sizeof(*elt) - sizeof(uuid_t));
                         return true; /* find tb policy */
                     }
                 }
 
                 elts_size += elt->size;
-                if ( elts_size >= pollist->policy_elements_size )
+                if ( elts_size >= all_elt_size )
                     break;
 
                 elt = (void *)elt + elt->size;
             }
-            if ( pollist->version == LCP_TPM12_POLICY_LIST_VERSION )
-                pollist = (void *)pollist + get_tpm12_policy_list_size(pollist);
-            else if ( pollist->version == LCP_TPM20_POLICY_LIST_VERSION )
-                pollist = (void *)pollist + get_tpm20_policy_list_size(
-                        (lcp_policy_list_t2 *)pollist);
+            //If necessary go to next list.
+            if ( MAJOR_VER(list_ver) == MAJOR_VER(LCP_TPM12_POLICY_LIST_VERSION)) {
+                pollist = (void *)pollist +
+                       get_tpm12_policy_list_size((lcp_policy_list_t *)pollist);
+            }
+            else if ( MAJOR_VER(list_ver) == MAJOR_VER(LCP_TPM20_POLICY_LIST_VERSION)) {
+                pollist = (void *)pollist + 
+                      get_tpm20_policy_list_size((lcp_policy_list_t2 *)pollist);
+            }
+            else if ( MAJOR_VER(list_ver) == MAJOR_VER(LCP_TPM20_POLICY_LIST2_1_VERSION_300) ){
+                pollist = (void *)pollist +
+                   get_raw_tpm20_list_2_1_size(&pollist->tpm20_policy_list_2_1);
+            }
         }
     }
 
@@ -337,8 +421,9 @@ tb_error_t set_policy(void)
     /* try to read tboot policy from TB_POLICY_INDEX in TPM NV */
     size_t policy_index_size = sizeof(_policy_index_buf);
     printk(TBOOT_INFO"reading Verified Launch Policy from TPM NV...\n");
-    if ( read_policy_from_tpm(tpm->tb_policy_index,
-             _policy_index_buf, &policy_index_size) ) {
+    if (read_policy_from_tpm(tpm->tb_policy_index, _policy_index_buf,
+                             &policy_index_size) &&
+            check_index_attribute(tpm->tb_policy_index)) {
         printk(TBOOT_DETA"\t:%lu bytes read\n", policy_index_size);
         if ( verify_policy((tb_policy_t *)_policy_index_buf,
                  policy_index_size, true) ) {
@@ -353,12 +438,13 @@ tb_error_t set_policy(void)
      * type is LCP_POLTYPE_LIST (since we could have been give a policy data
      * file even though the policy was not a LIST */
     printk(TBOOT_INFO"reading Launch Control Policy from TPM NV...\n");
-    if ( read_policy_from_tpm(tpm->lcp_own_index,
-             _policy_index_buf, &policy_index_size) ) {
+    if (read_policy_from_tpm(tpm->lcp_own_index, _policy_index_buf,
+                             &policy_index_size) &&
+            check_index_attribute(tpm->lcp_own_index)) {
         printk(TBOOT_DETA"\t:%lu bytes read\n", policy_index_size);
         /* assume lcp policy has been verified by sinit already */
         lcp_policy_t *pol = (lcp_policy_t *)_policy_index_buf;
-        if ( pol->version == LCP_DEFAULT_POLICY_VERSION_2 &&
+        if ( MAJOR_VER(pol->version) == MAJOR_VER(LCP_DEFAULT_POLICY_VERSION_2) &&
              pol->policy_type == LCP_POLTYPE_LIST && unwrap_lcp_policy() ) {
             if ( verify_policy((tb_policy_t *)_policy_index_buf,
                      calc_policy_size((tb_policy_t *)_policy_index_buf),
@@ -366,7 +452,8 @@ tb_error_t set_policy(void)
                 goto policy_found;
         }
         lcp_policy_t2 *pol2 = (lcp_policy_t2 *)_policy_index_buf;
-        if ( pol2->version == LCP_DEFAULT_POLICY_VERSION &&
+        /*Only check if major version is 3, minor is checked by sinit*/
+        if ( MAJOR_VER(pol2->version) == MAJOR_VER(LCP_DEFAULT_POLICY_VERSION) &&
              pol2->policy_type == LCP_POLTYPE_LIST && unwrap_lcp_policy() ) {
             if ( verify_policy((tb_policy_t *)_policy_index_buf,
                      calc_policy_size((tb_policy_t *)_policy_index_buf),
@@ -378,7 +465,11 @@ tb_error_t set_policy(void)
 
     /* either no policy in TPM NV or policy is invalid, so use default */
     printk(TBOOT_WARN"failed to read policy from TPM NV, using default\n");
-    g_policy = g_using_da ? &_def_policy_da : &_def_policy;
+    if (g_tpm_ver == TPM_VER_12) {
+        g_policy = &_def_policy_12;
+    } else {
+        g_policy = g_using_da ? &_def_policy_da : &_def_policy;
+    }
     policy_index_size = calc_policy_size(g_policy);
 
     /* sanity check; but if it fails something is really wrong */
@@ -439,7 +530,7 @@ static bool hash_module(hash_list_t *hl,
         hl->count = 1;
         hl->entries[0].alg = tpm->cur_alg;
 
-        if ( !hash_buffer((const unsigned char *)cmdline, strlen(cmdline),
+        if ( !hash_buffer((const unsigned char *)cmdline, tb_strlen(cmdline),
                     &hl->entries[0].hash, tpm->cur_alg) )
             return false;
         /* hash image and extend into cmdline hash */
@@ -455,10 +546,10 @@ static bool hash_module(hash_list_t *hl,
     {
         hash_list_t img_hl, final_hl;
         if ( !tpm_fp->hash(tpm, 2, (const unsigned char *)cmdline,
-                strlen(cmdline), hl) )
+                tb_strlen(cmdline), hl) )
             return false;
 
-        uint8_t buf[128];
+        uint8_t buf[2*sizeof(tb_hash_t)];
 
         if ( !tpm_fp->hash(tpm, 2, base, size, &img_hl) )
             return false;
@@ -496,7 +587,7 @@ static bool hash_module(hash_list_t *hl,
         hl->count = tpm->alg_count;
         for (unsigned int i=0; i<hl->count; i++) {
             hl->entries[i].alg = tpm->algs[i];
-            if ( !hash_buffer((const unsigned char *)cmdline, strlen(cmdline),
+            if ( !hash_buffer((const unsigned char *)cmdline, tb_strlen(cmdline),
                         &hl->entries[i].hash, tpm->algs[i]) )
                 return false;
 
@@ -629,12 +720,12 @@ static tb_error_t verify_module(module_t *module, tb_policy_entry_t *pol_entry,
     if ( pol_entry != NULL ) {
         /* chunk the command line into 80 byte chunks */
 #define CHUNK_SIZE 80
-        int      cmdlen = strlen(cmdline);
+        int      cmdlen = tb_strlen(cmdline);
         char    *cptr = cmdline;
         char     cmdchunk[CHUNK_SIZE+1];
         printk(TBOOT_INFO"verifying module \"");
         while (cmdlen > 0) {
-            strncpy(cmdchunk, cptr, CHUNK_SIZE);
+            tb_strncpy(cmdchunk, cptr, CHUNK_SIZE);
             cmdchunk[CHUNK_SIZE] = 0;
             printk(TBOOT_INFO"\n%s", cmdchunk);
             cmdlen -= CHUNK_SIZE;
@@ -690,8 +781,8 @@ static void verify_g_policy(void)
     /* hash will be <policy control field (4bytes)> | <hash policy> */
     /* where <hash policy> will be 0s if TB_POLCTL_EXTEND_PCR17 is clear */
     static uint8_t buf[sizeof(tb_hash_t) + sizeof(g_policy->policy_control)];
-    memset(buf, 0, sizeof(buf));
-    memcpy(buf, &g_policy->policy_control, sizeof(g_policy->policy_control));
+    tb_memset(buf, 0, sizeof(buf));
+    tb_memcpy(buf, &g_policy->policy_control, sizeof(g_policy->policy_control));
     if ( g_policy->policy_control & TB_POLCTL_EXTEND_PCR17 ) {
         if ( !hash_policy((tb_hash_t *)&buf[sizeof(g_policy->policy_control)],
                           tpm->cur_alg) ) {
@@ -748,6 +839,11 @@ static void verify_g_policy(void)
 
 void verify_all_modules(loader_ctx *lctx)
 {
+    if (!verify_loader_context(lctx)) {
+        printk(TBOOT_ERR"Error: Invalid loader context\n");
+        apply_policy(TB_ERR_FATAL);
+    }
+
     /* assumes mbi is valid */
     verify_g_policy();
 
@@ -839,7 +935,7 @@ static tb_error_t verify_nvindex(tb_policy_entry_t *pol_entry,
     }
 
     /* get nv content */
-    memset(nv_buf, 0, sizeof(nv_buf));
+    tb_memset(nv_buf, 0, sizeof(nv_buf));
     if ( !read_policy_from_tpm(pol_entry->nv_index,
                 nv_buf, &nv_size) ) {
         printk(TBOOT_ERR"\t :reading nv index 0x%08X failed\n",
@@ -862,7 +958,7 @@ static tb_error_t verify_nvindex(tb_policy_entry_t *pol_entry,
                    (int)nv_size, sizeof(digest.sha1));
             return TB_ERR_NV_VERIFICATION_FAILED;
         }
-        memcpy(digest.sha1, nv_buf, nv_size);
+        tb_memcpy(digest.sha1, nv_buf, nv_size);
         break;
     default:
         printk(TBOOT_ERR"\t :bad mod number for NV measuring in policy entry: %d\n",
@@ -879,7 +975,7 @@ static tb_error_t verify_nvindex(tb_policy_entry_t *pol_entry,
         VL_ENTRIES(NUM_VL_ENTRIES).pcr = pol_entry->pcr;
         VL_ENTRIES(NUM_VL_ENTRIES).hl.count = 1;
         VL_ENTRIES(NUM_VL_ENTRIES).hl.entries[0].alg = TB_HALG_SHA1;
-        memcpy(VL_ENTRIES(NUM_VL_ENTRIES++).hl.entries[0].hash.sha1,
+        tb_memcpy(VL_ENTRIES(NUM_VL_ENTRIES++).hl.entries[0].hash.sha1,
                 digest.sha1, SHA1_LENGTH);
     }
 
